# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
import logging
from pprint import pprint
import pytest  # noqa import pytest

from crush.main import Main
from crush import Crush

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


class TestCompare(object):

    def setup_class(self):
        pass

    def define_crushmap_10(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "indep": [
                    ["take", "dc1"],
                    ["chooseleaf", "indep", 0, "type", "host"],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 1},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 2},
                ],
            } for i in range(0, 10)
        ])
        return crushmap

    def define_crushmaps_1(self):
        crushmap = self.define_crushmap_10()
        pprint(crushmap)
        c1 = Crush()
        c1.parse(crushmap)

        m2 = copy.deepcopy(crushmap)
        del m2['trees'][0]['children'][2]['children'][1]
        c2 = Crush()
        c2.parse(m2)

        return (c1, c2)

    def test_display(self):
        c1, c2 = self.define_crushmaps_1()
        c = Main().constructor([
            'compare',
            '--rule', 'indep',
            '--replication-count', '2',
            '--values-count', '10',
        ])
        c.set_origin(c1)
        c.set_destination(c2)
        c.compare()
        out = c.display()
        print(out)
        assert 'device04        0        0        1        1   10.00%' in out
        assert 'objects%   10.00%    5.00%    5.00%    5.00%   25.00%' in out

    def test_compare_bucket_firstn(self):
        origin = self.define_crushmap_10()
        pprint(origin)

        # firstn, mapping order does not matter
        c = Main().constructor([
            '--verbose',
            'compare',
            '--rule', 'firstn',
            '--replication-count', '2',
        ])

        c.set_origin_crushmap(origin)

        #
        # Swapping weights within a bucket, values stay
        # in the bucket but move between children
        #
        destination = copy.deepcopy(origin)
        host0 = destination['trees'][0]['children'][0]
        w0 = host0['children'][0]['weight']
        w1 = host0['children'][1]['weight']
        host0['children'][0]['weight'] = w1
        host0['children'][1]['weight'] = w0
        c.set_destination_crushmap(destination)

        c.args.values_count = 100
        (from_to, in_out) = c.compare_bucket(host0)
        assert from_to == {
            'device01': {'device00': 4}
        }
        assert in_out == {}

        #
        # Increasing the weight of the items it contains changes the
        # weight of the bucket and items move in/out of the bucket.
        #
        print("collisions")
        destination = copy.deepcopy(origin)
        host0 = destination['trees'][0]['children'][0]
        host0['children'][0]['weight'] *= 10
        c.set_destination_crushmap(destination)

        c.args.values_count = 10
        (from_to, in_out) = c.compare_bucket(host0)
        print("from_to " + str(from_to))
        print("in_out " + str(in_out))
        assert from_to == {
            'device01': {'device00': 1}
        }
        assert in_out == {
            'device05': {'device00': 1},
            'device08': {'device00': 1},
            'device04': {'device00': 1},
        }

    def test_compare_bucket_indep(self):
        origin = self.define_crushmap_10()

        # indep, mapping order does not matter
        c = Main().constructor([
            'compare',
            '--rule', 'indep',
            '--replication-count', '2',
            '--order-matters',
        ])

        c.set_origin_crushmap(origin)

        #
        # Swapping weights within a bucket, values stay
        # in the bucket but move between children
        #
        destination = copy.deepcopy(origin)
        host0 = destination['trees'][0]['children'][0]
        w0 = host0['children'][0]['weight']
        w1 = host0['children'][1]['weight']
        host0['children'][0]['weight'] = w1
        host0['children'][1]['weight'] = w0
        c.set_destination_crushmap(destination)

        c.args.values_count = 100
        (from_to, in_out) = c.compare_bucket(host0)
        assert from_to == {
            'device01': {'device00': 5}
        }
        assert in_out == {}

        #
        # Increasing the weight of the items it contains changes the
        # weight of the bucket and items move in/out of the bucket.
        #
        destination = copy.deepcopy(origin)
        host0 = destination['trees'][0]['children'][0]
        host0['children'][0]['weight'] *= 10
        c.set_destination_crushmap(destination)

        c.args.values_count = 30
        (from_to, in_out) = c.compare_bucket(host0)
        print("from_to " + str(from_to))
        print("in_out " + str(in_out))
        assert from_to == {
            'device01': {'device00': 1}
        }
        assert in_out == {
            'device01': {'device05': 1},
            'device04': {'device00': 1},
            'device05': {'device00': 2},
            'device09': {'device01': 1},
            'device13': {'device00': 1},
            'device17': {'device00': 1},
            'device19': {'device00': 1},
            'device00': {'device19': 1},
        }

    def test_compare(self):
        c1, c2 = self.define_crushmaps_1()

        # firstn, mapping order does not matter
        c = Main().constructor([
            'compare',
            '--rule', 'firstn',
            '--replication-count', '2',
            '--values-count', '10',
        ])

        # device05 is removed
        c.set_origin(c1)
        c.set_destination(c2)
        assert c.compare() == {
            'device04': {'device17': 1},
            'device05': {'device12': 1, 'device04': 2, 'device08': 1}
        }
        # device05 is added
        c.set_origin(c2)
        c.set_destination(c1)
        assert c.compare() == {
            'device12': {'device05': 1},
            'device17': {'device04': 1},
            'device04': {'device05': 2},
            'device08': {'device05': 1}
        }

        # indep, mapping order matters
        c = Main().constructor([
            'compare',
            '--rule', 'indep',
            '--replication-count', '2',
            '--values-count', '10',
            '--order-matters',
        ])
        # device05 is removed
        c.set_origin(c1)
        c.set_destination(c2)
        assert c.compare() == {
            'device04': {'device13': 1, 'device17': 1},
            'device05': {'device04': 2, 'device08': 1}
        }
        # device05 is added
        c.set_origin(c2)
        c.set_destination(c1)
        assert c.compare() == {
            'device13': {'device04': 1},
            'device17': {'device04': 1},
            'device04': {'device05': 2},
            'device08': {'device05': 1}
        }

    def define_crushmaps_2(self):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "indep": [
                    ["take", "dc1"],
                    ["chooseleaf", "indep", 0, "type", "host"],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 50},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 50},
                ],
            } for i in range(0, 5)
        ])
        first = crushmap['trees'][0]['children'][0]['children']
        first[0]['weight'] = 5
        first[1]['weight'] = 5
        pprint(crushmap)
        c1 = Crush()
        c1.parse(crushmap)

        m2 = copy.deepcopy(crushmap)
        del m2['trees'][0]['children'][4]
        c2 = Crush()
        c2.parse(m2)

        return (c1, c2)

    def test_compare_display(self):
        c = Main().constructor([
            'compare',
            '--rule', 'firstn',
            '--values-count', '1000',
        ])
        c1, c2 = self.define_crushmaps_2()
        c.set_origin(c2)
        c.set_destination(c1)
        c.args.replication_count = 1
        c.compare()
        out = c.display()
        print(out)
        assert "objects%   10.20%   12.70%   22.90%" in out
        c.args.replication_count = 3
        c.compare()
        out = c.display()
        print(out)
        assert ("objects%    0.17%    0.13%    0.97%    0.77%    0.77%"
                "    0.63%    0.73%    0.87%   12.30%   12.20%   29.53%") in out

    def test_origin_weights(self):
        a = Main().constructor([
            "compare", "--rule", "replicated_ruleset",
            "--replication-count", "1",
            "--origin", "tests/weights-crushmap.json",
            "--destination", "tests/weights-crushmap.json",
            "--origin-weights", "tests/weights.json"])
        a.args.backward_compatibility = True
        a.run_compare()

    def test_destination_weights(self):
        a = Main().constructor([
            "compare", "--rule", "replicated_ruleset",
            "--replication-count", "1",
            "--origin", "tests/weights-crushmap.json",
            "--destination", "tests/weights-crushmap.json",
            "--destination-weights", "tests/weights.json"])
        a.args.backward_compatibility = True
        a.run_compare()

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -vv -s tests/test_compare.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import argparse
import collections
import logging
import textwrap

from crush import Crush
from crush import analyze
from crush import compare
from crush import optimize

log = logging.getLogger('crush')


class Main(object):

    def __init__(self):
        self.create_parser()

        if self.parser.get_default('backward_compatibility') is None:
            self.parser.set_defaults(backward_compatibility=False)

        self.parser.add_argument(
            '-v', '--verbose',
            action='store_true', default=None,
            help='be more verbose',
        )

        self.parser.add_argument(
            '--debug',
            action='store_true', default=None,
            help='debugging output, very verbose',
        )

        self.subparsers = self.parser.add_subparsers(
            title='subcommands',
            description='valid subcommands',
            help='sub-command -h',
        )

        analyze.Analyze.set_parser(self.subparsers, self.hook_analyze_args)
        compare.Compare.set_parser(self.subparsers, self.hook_compare_args)
        optimize.Optimize.set_parser(self.subparsers, self.hook_optimize_args)

    def create_parser(self):
        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            A library to control placement in a hierarchy
            """))

    def __getstate__(self):
        return (self.argv,)

    def __setstate__(self, state):
        self.__init__()
        self.parse(state[0])

    def clone(self):
        return Main()

    def parse(self, argv):
        self.argv = argv
        self.args = self.parser.parse_args(argv)

        if self.args.debug:
            level = logging.DEBUG
            format = '%(asctime)s %(funcName)20s %(message)s'
        elif self.args.verbose:
            level = logging.INFO
            format = '%(asctime)s %(funcName)20s %(message)s'
        else:
            level = logging.WARNING
            format = '%(asctime)s %(message)s'
        if log.getEffectiveLevel() == 0 or log.getEffectiveLevel() > level:
            log.setLevel(level)
        logging.basicConfig(format=format)

    @staticmethod
    def get_trimmed_argv(to_parser, args):
        options = []
        positionals = []
        dest2option = {}
        known_dests = []
        for action in to_parser._actions:
            option = list(filter(lambda o: o in to_parser._option_string_actions,
                                 action.option_strings))
            if len(option) > 0:
                dest2option[action.dest] = option[0]
            known_dests.append(action.dest)
        v = collections.OrderedDict(sorted(vars(args).items(),
                                           key=lambda t: t[0]))
        for (key, value) in v.items():
            log.debug('get_trimmed_argv: checking ' +
                      str(key) + "=" + str(value))
            if key not in known_dests:
                log.debug('get_trimmed_argv: skip unknown ' + str(key))
                continue
            if key in dest2option:
                option = dest2option[key]
                action = to_parser._option_string_actions[option]
                if value != action.default:
                    if action.nargs is None or action.nargs == 1:
                        options.extend([option, str(value)])
                    elif action.nargs == 0:
                        options.append(option)
            else:
                log.debug('get_trimmed_argv: positional ' +
                          str(key) + "=" + str(value))
                positionals.extend(value)
        return options + positionals

    def constructor(self, argv):
        self.parse(argv)
        return self.args.func(self.args, self)

    def main(self, argv):
        return self.constructor(argv).run()

    def hook_analyze_args(self, parser):
        pass

    def hook_compare_args(self, parser):
        pass

    def hook_optimize_args(self, parser):
        pass

    def hook_create_values(self):
        values = range(0, self.args.values_count)
        return dict(zip(values, values))

    def value_name(self):
        return 'objects'

    def convert_to_crushmap(self, crushmap):
        c = Crush(verbose=self.args.debug,
                  backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        return c.get_crushmap()

    def crushmap_to_file(self, crushmap):
        c = Crush(verbose=self.args.debug,
                  backward_compatibility=self.args.backward_compatibility)
        c.parse(crushmap)
        c.to_file(self.args.out_path)



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import copy
import logging
import pytest # noqa needed for capsys
import os
import pickle

from crush import Crush
from crush.ceph import Ceph
from crush.main import Main
from crush.analyze import BadMapping

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)


class TestOptimize(object):

    def test_pickle(self):
        o = Ceph().constructor(['optimize'])
        p = pickle.dumps(o)
        oo = pickle.loads(p)
        assert oo.main.argv == o.main.argv
        assert type(oo) == type(o)

    def test_set_optimize_args(self):
        a = Ceph().constructor([
            'optimize',
            '--pool', '3',
        ])
        a.args.replication_count = None
        assert a.args.choose_args is None
        assert a.args.rule is None
        assert a.args.pg_num is None
        assert a.args.pgp_num is None
        assert a.args.out_version == 'luminous'
        assert a.args.with_positions is True
        a.main.convert_to_crushmap('tests/ceph/ceph-report-small.json')
        assert a.args.replication_count == 3
        assert a.args.choose_args == '3'
        assert a.args.rule == 'data'
        assert a.args.pg_num == 1
        assert a.args.pgp_num == 1
        assert a.args.out_version == 'j'
        assert a.args.with_positions is False

    def test_get_choose_arg(self):
        a = Ceph().constructor([
            'optimize',
            '--choose-args', 'optimize',
        ])
        crushmap = {}

        bucket = {'id': -1}

        choose_arg = a.get_choose_arg(crushmap, bucket)
        assert {'bucket_id': -1} == choose_arg
        assert [choose_arg] == crushmap['choose_args']['optimize']

        choose_arg = a.get_choose_arg(crushmap, bucket)
        assert {'bucket_id': -1} == choose_arg
        assert [choose_arg] == crushmap['choose_args']['optimize']

        bucket = {'id': -2}

        choose_arg = a.get_choose_arg(crushmap, bucket)
        assert [{'bucket_id': -1}, {'bucket_id': -2}] == crushmap['choose_args']['optimize']

    def test_set_choose_arg_position(self):
        a = Ceph().constructor([
            'optimize',
            '--choose-args', 'optimize',
        ])
        bucket = {
            'id': -1,
            'children': [
                {'id': 1, 'weight': 10},
                {'id': 2, 'weight': 20},
            ],
        }
        choose_arg = {'bucket_id': -1}
        a.set_choose_arg_position(choose_arg, bucket, 0)
        assert {'bucket_id': -1, 'weight_set': [[10, 20]]} == choose_arg

        a.set_choose_arg_position(choose_arg, bucket, 1)
        assert {'bucket_id': -1, 'weight_set': [[10, 20], [10, 20]]} == choose_arg

        choose_arg['weight_set'][1] = [100, 200]
        a.set_choose_arg_position(choose_arg, bucket, 3)
        expected = {'bucket_id': -1, 'weight_set': [
            [10, 20], [100, 200], [100, 200], [100, 200]
        ]}
        assert expected == choose_arg

        a.set_choose_arg_position(choose_arg, bucket, 1)
        assert expected == choose_arg

    def make_bucket(self, weights):
        crushmap = {
            "choose_args": {"optimize": []},
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "weight": sum(weights),
                    "children": [],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["set_choose_tries", 100],
                    ["choose", "firstn", 0, "type", 0],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "id": i,
                "name": "device%d" % i,
                "weight": weights[i] * 0x10000,
            } for i in range(len(weights))
        ])
        return crushmap

    def verify_optimize(self, weights, expected_delta,
                        values_count, replication_count):
        crushmap = self.make_bucket(weights)
        rule = 'firstn'
        p = [
            '--values-count', str(values_count),
            '--replication-count', str(replication_count),
            '--rule', rule,
            '--choose-args', 'optimize',
        ]
        o = Main().constructor(['--verbose', 'optimize'] + p)
        origin_crushmap = copy.deepcopy(crushmap)
        bucket = crushmap['trees'][0]
        previous_weight_set = []
        for position in range(replication_count):
            o.optimize_replica(
                p, origin_crushmap, crushmap, bucket, replication_count, position)
            assert 1 == len(crushmap['choose_args']['optimize'])
            weight_set = copy.deepcopy(crushmap['choose_args']['optimize'][0]['weight_set'])
            assert weight_set[:len(previous_weight_set)] == previous_weight_set

        a = Main().constructor(['analyze', '--choose-args', 'optimize'] + p)
        r = a.analyze_crushmap(crushmap)
        delta = (r['~objects~'] - r['~expected~']).tolist()
        print("delta = " + str(delta))
        assert expected_delta == delta

    def test_simple(self):
        # values_count=100 is not enough samples
        self.verify_optimize([1, 1, 1], [0, 0, 0, 0],
                             values_count=100, replication_count=2)

        self.verify_optimize([5, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                             [0, -3, 0, 1, 0, 0, 0, 0, 0, 1, 1],
                             values_count=100, replication_count=6)

        self.verify_optimize([1, 2, 3, 1, 2, 3, 1, 2, 3, 1],
                             [0, 0, 0, 0, -1, 0, 0, 0, 1, 0, 0],
                             values_count=100, replication_count=6)

    def test_very_different_weights(self):
        # values_count=10000 is not enough samples
        delta = [0, -1, -1, -1, -1, 1, -1, -1, -1, 0, -1, 1, 0, -1, -1, 1, 2, 1, 2, 1, 0, 0, 1,
                 -1, -1, -1, 0, 0, 1, 1, 0, -1, 0, 0, 1, 1, -1, -1, -1, -1, 0, 0, 1, 1, -1, 0,
                 0, 1, 0, -1, -1, 0, 2, 1, 0, -1, 1, 0, 0, -1, 0, 0, 2, -1, 1, 1, 1, -1, 0, 1,
                 0, -1, 0, 0, 1, -1, 1, -1, 0, 1, 0, 0, 1, -1, 1, -1, 1, -1, -1, 1, -1, 1, 0,
                 -1, 0, 0, 0, 0, 0, -1, 0]
        self.verify_optimize(range(1, 101), delta,
                             values_count=10000, replication_count=1)

    def test_overweighted(self):
        # 5 is overweighted for replica > 1 because
        self.verify_optimize([5, 1, 1, 1, 1], [0, 0, 0, 0, 0, 0],
                             values_count=100, replication_count=1)
        self.verify_optimize([5, 1, 1, 1, 1], [0, -3, 0, 0, 1, 2],
                             values_count=100, replication_count=2)

    @pytest.mark.skipif(os.environ.get('LONG') is None, reason="LONG")
    def test_probability_bias(self):
        self.verify_optimize([5, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                             [0, 0, 0, 0, 0, 0, -1, -1, 1, 0, 1],
                             replication_count=2,
                             values_count=100000)

    def test_bad_mapping(self):
        with pytest.raises(BadMapping):
            self.verify_optimize([1000, 1, 1, 1, 1], [],
                                 replication_count=2, values_count=100)

    def run_optimize(self, p, crushmap, gain):
        o = Ceph().constructor(['optimize', '--choose-args', 'optimize'] + p)
        origin_crushmap = copy.deepcopy(crushmap)
        (count, optimized) = o.optimize(crushmap)
        self.analyze_optimization(p, origin_crushmap, optimized, gain)

    def analyze_optimization(self, p, crushmap, optimized, gain):
        a = Ceph().constructor(['analyze'] + p)
        before = a.analyze_crushmap(crushmap)
        print("============= before")
        print(str(before))

        a = Ceph().constructor(['analyze', '--choose-args', 'optimize'] + p)
        after = a.analyze_crushmap(optimized)
        print("============= after")
        print(str(after))

        for type in before['~type~'].unique():
            b = before.loc[before['~type~'] == type]
            b = b.sort_values(by='~over/under filled %~', ascending=False)
            b_span = b.iloc[0]['~over/under filled %~'] - b.iloc[-1]['~over/under filled %~']
            a = after.loc[after['~type~'] == type]
            a = a.sort_values(by='~over/under filled %~', ascending=False)
            a_span = a.iloc[0]['~over/under filled %~'] - a.iloc[-1]['~over/under filled %~']
            print("============= span " + str(type) + " before " +
                  str(b_span) + " after " + str(a_span))
            assert a_span <= b_span / gain

    def test_optimize_one_step(self):
        pg_num = 2048
        size = 3
        a = Ceph().constructor([
            'optimize',
            '--no-multithread',
            '--replication-count', str(size),
            '--pool', '3',
            '--pg-num', str(pg_num),
            '--pgp-num', str(pg_num),
            '--rule', 'data',
            '--choose-args', 'optimize',
            '--step', '64',
        ])
        c = Crush(backward_compatibility=True)
        c.parse('tests/test_optimize_small_cluster.json')
        crushmap = c.get_crushmap()
        (count, crushmap) = a.optimize(crushmap)
        assert 240 == count

    def test_optimize_report_compat(self):
        #
        # verify --choose-args is set to --pool when the crushmap contains
        # *-target-weights buckets.
        #
        expected_path = 'tests/ceph/ceph-report-compat-optimized.txt'
        out_path = expected_path + ".err"
        for p in ([],
                  ['--choose-args=3'],
                  ['--pool=3'],
                  ['--choose-args=3', '--pool=3']):
            Ceph().main([
                '--verbose',
                'optimize',
                '--no-multithread',
                '--crushmap', 'tests/ceph/ceph-report-compat.json',
                '--out-path', out_path,
                '--out-format', 'txt',
            ] + p)
            assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
            os.unlink(out_path)

    @pytest.mark.skipif(os.environ.get('LONG') is None, reason="LONG")
    def test_optimize_small_cluster(self):
        pg_num = 4096
        size = 3
        p = [
            '--replication-count', str(size),
            '--pool', '3',
            '--pg-num', str(pg_num),
            '--pgp-num', str(pg_num),
            '--rule', 'data',
        ]
        self.run_optimize(p, 'tests/test_optimize_small_cluster.json', 10)

    @pytest.mark.skipif(os.environ.get('LONG') is None, reason="LONG")
    def test_optimize_big_cluster(self):
        # few samples
        pg_num = 2048
        size = 3
        p = [
            '--replication-count', str(size),
            '--pool', '5',
            '--pg-num', str(pg_num),
            '--pgp-num', str(pg_num),
            '--rule', 'data',
        ]
        self.run_optimize(p, 'tests/test_optimize_big_cluster.json', 4)

    @pytest.mark.skipif(os.environ.get('LONG') is None, reason="LONG")
    def test_optimize_step(self):
        pg_num = 2048
        size = 3
        a = Ceph().constructor([
            'optimize',
            '--no-multithread',
            '--replication-count', str(size),
            '--pool', '3',
            '--pg-num', str(pg_num),
            '--pgp-num', str(pg_num),
            '--rule', 'data',
            '--choose-args', 'optimize',
            '--step', '64',
        ])
        c = Crush(backward_compatibility=True)
        c.parse('tests/test_optimize_small_cluster.json')
        crushmap = c.get_crushmap()
        converged = False
        for i in range(20):
            (count, crushmap) = a.optimize(crushmap)
            if count <= 0:
                converged = True
                break
            print("moved " + str(count) + " values")
        assert converged

    @pytest.mark.skipif(os.environ.get('LONG') is None, reason="LONG")
    def test_optimize_step_forecast(self, caplog):
        expected_path = 'tests/test_optimize_small_cluster_step_1.txt'
        out_path = expected_path + ".err"
        # few samples
        pg_num = 2048
        size = 3
        Ceph().main([
            '--verbose',
            'optimize',
            '--no-multithread',
            '--crushmap', 'tests/test_optimize_small_cluster.json',
            '--out-path', out_path,
            '--out-format', 'txt',
            '--replication-count', str(size),
            '--pool', '2',
            '--pg-num', str(pg_num),
            '--pgp-num', str(pg_num),
            '--rule', 'data',
            '--choose-args', '0',
            '--step', '64',
        ])

        assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
        os.unlink(out_path)
        assert 'step 2 moves 73 objects' in caplog.text()
        assert 'step 3 moves 76 objects' in caplog.text()
        assert 'step 4 moves 93 objects' in caplog.text()
        assert 'step 5 moves 80 objects' in caplog.text()
        assert 'step 6 moves 100 objects' in caplog.text()
        assert 'step 7 moves 80 objects' in caplog.text()
        assert 'step 8 moves 53 objects' in caplog.text()
        assert 'step 9 moves 0 objects' in caplog.text()

# Local Variables:
# compile-command: "cd .. ; LONG=yes tox -e py27 -- -s -vv tests/test_optimize.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import division

import argparse
import logging
import textwrap


log = logging.getLogger(__name__)


class Convert(object):

    def __init__(self, args, main):
        self.args = args
        self.main = main

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(
            add_help=False,
            conflict_handler='resolve',
        )
        parser.add_argument(
            '--out-path',
            help='path of the output file')
        return parser

    @staticmethod
    def set_parser(subparsers, arguments):
        parser = Convert.get_parser()
        arguments(parser)
        subparsers.add_parser(
            'convert',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent("""\
            The Ceph crushmap can be stored in three formats:

            - JSON, the output of *ceph osd crush dump* or *ceph report*

            - txt, the output of *crushtool -d*

            - binary, the output of *crushtool -c* or *ceph getcrushmap*

            The JSON used is different from the python-crush format documented at
            http://crush.readthedocs.io/en/latest/api.html#crush.Crush.parse.

            It supports any of the existing Ceph formats and is
            compatible with Luminous and below. It converts the
            crushmap into the python-crush format and display the
            result on the standard output.

            """),
            epilog=textwrap.dedent("""
            Examples:

            Convert a Ceph JSON crushmap into a python-crush crushmap:
            - crush convert --in-path crushmap-ceph.json --out-path crushmap.json

            Convert a Ceph text crushmap into a python-crush crushmap:
            - crush convert --in-path crushmap.txt --out-path crushmap.json

            Convert a binary crushmap to python-crush crushmap:
            - crush convert --in-path crushmap.bin --out-path crushmap.json

            Convert a python-crush crushmap to Ceph text crushmap
            - crush convert --in-path crushmap.json \\
                            --out-path crushmap.json --out-format txt
            """),
            help='Convert crushmaps',
            parents=[parser],
        ).set_defaults(
            func=Convert,
        )

    def run(self):
        crushmap = self.main.convert_to_crushmap(self.args.in_path)
        self.main.crushmap_to_file(crushmap)



import json
from crush import Crush

crushmap = """
{
  "trees": [
    {
      "type": "root", "name": "dc1", "id": -1,
      "children": [
        {
         "type": "host", "name": "host0", "id": -2,
         "children": [
          { "id": 0, "name": "device0", "weight": 65536 },
          { "id": 1, "name": "device1", "weight": 131072 }
         ]
        },
        {
         "type": "host", "name": "host1", "id": -3,
         "children": [
          { "id": 2, "name": "device2", "weight": 65536 },
          { "id": 3, "name": "device3", "weight": 131072 }
         ]
        },
        {
         "type": "host", "name": "host2", "id": -4,
         "children": [
          { "id": 4, "name": "device4", "weight": 65536 },
          { "id": 5, "name": "device5", "weight": 131072 }
         ]
        }
      ]
    }
  ],
  "rules": {
    "data": [
      [ "take", "dc1" ],
      [ "chooseleaf", "firstn", 0, "type", "host" ],
      [ "emit" ]
    ]
  }
}

"""

c = Crush()
c.parse(json.loads(crushmap))
print(c.map(rule="data", value=1234, replication_count=1))
print(c.map(rule="data", value=1234, replication_count=2))



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import json
import pytest  # noqa needed for capsys

from crush.ceph import Ceph
from crush import ceph

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


class TestCephReport(object):

    def load(self):
        with open('tests/ceph/ceph-report-small.json') as f:
            return json.load(f)

    def test_good_by_default(self):
        ceph.CephReport().parse_report(self.load())

    def test_fail_health(self):
        report = self.load()
        report['health']['overall_status'] = 'HEALTH_WARN'
        with pytest.raises(ceph.HealthError):
            ceph.CephReport().parse_report(report)

    def test_fail_primary_affinity(self):
        report = self.load()
        report['osdmap']['osds'][0]['primary_affinity'] = 0.5
        with pytest.raises(ceph.UnsupportedError) as e:
            ceph.CephReport().parse_report(report)
        assert 'affinity is !=' in str(e.value)

    def test_fail_pool_type(self):
        report = self.load()
        report['osdmap']['pools'][0]['type'] = 2
        with pytest.raises(ceph.UnsupportedError) as e:
            ceph.CephReport().parse_report(report)
        assert 'only type == 1' in str(e.value)

    def test_fail_pool_object_hash(self):
        report = self.load()
        report['osdmap']['pools'][0]['object_hash'] = 5
        with pytest.raises(ceph.UnsupportedError) as e:
            ceph.CephReport().parse_report(report)
        assert 'only object_hash == 2' in str(e.value)

    def test_fail_pool_flags_names(self):
        report = self.load()
        report['osdmap']['pools'][0]['flags_names'] = 'somethingelse'
        with pytest.raises(ceph.UnsupportedError) as e:
            ceph.CephReport().parse_report(report)
        assert 'only hashpspool' in str(e.value)

    def test_fail_mapping_name(self, caplog):
        report = self.load()
        report['pgmap']['pg_stats'][0]['pgid'] = '5.1'
        with pytest.raises(ceph.MappingError):
            ceph.CephReport().parse_report(report)
        assert 'is not in pgmap' in caplog.text()

    def test_fail_mapping_osds(self, caplog):
        report = self.load()
        report['pgmap']['pg_stats'][0]['acting'] = [1, 2, 3]
        with pytest.raises(ceph.MappingError):
            ceph.CephReport().parse_report(report)
        assert 'instead of [1, 2, 3]' in caplog.text()


class TestCeph(object):

    def test_conversions(self):
        base = 'tests/sample-ceph-crushmap.'
        for ext_in in ('txt', 'crush', 'json', 'python-json'):
            in_path = base + ext_in
            for ext_out in ('txt', 'crush', 'json', 'python-json'):
                expected_path = base + ext_out
                out_path = expected_path + ".err"
                print("conversion " + in_path + " => " + expected_path)
                Ceph().main([
                    'convert',
                    '--in-path', in_path,
                    '--out-path', out_path,
                    '--out-format', ext_out,
                ])
                if ext_out == 'crush':
                    cmd = "cmp"
                else:
                    cmd = "diff -Bbu"
                assert os.system(cmd + " " + expected_path + " " + out_path) == 0
                os.unlink(out_path)

    def test_report(self):
        in_path = 'tests/ceph/ceph-report.json'
        expected_path = 'tests/ceph/crushmap-from-ceph-report.json'
        out_path = expected_path + ".err"
        Ceph().main([
            'convert',
            '--in-path', in_path,
            '--out-path', out_path,
            '--out-format', 'python-json',
        ])
        assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
        os.unlink(out_path)

    def test_report_compat(self):
        #
        # verify --choose-args is set to the pool when the crushmap contains
        # *-target-weights buckets.
        #
        expected_path = 'tests/ceph/ceph-report-compat-converted.txt'
        out_path = expected_path + ".err"
        Ceph().main([
            '--verbose',
            'convert',
            '--in-path', 'tests/ceph/ceph-report-compat.json',
            '--out-path', out_path,
            '--out-format', 'txt',
        ])
        assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
        os.unlink(out_path)

    def test_rules_order(self):
        expected_path = 'tests/ceph/ceph-crushmap-rules-order.txt'
        out_path = expected_path + ".err"
        Ceph().main([
            '--verbose',
            'convert',
            '--in-path', 'tests/ceph/ceph-crushmap-rules-order.json',
            '--out-path', out_path,
            '--out-format', 'txt',
        ])
        assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
        os.unlink(out_path)

    def test_hook_create_values(self):
        c = Ceph()
        c.parse([
            '--verbose',
            'analyze',
            '--values-count', '2',
        ])
        assert {0: 0, 1: 1} == c.hook_create_values()
        c.parse([
            '--verbose',
            'analyze',
            '--pool', '2',
            '--pg-num', '3',
            '--pgp-num', '3',
        ])
        expected = {u'2.0': -113899774, u'2.1': -1215435108, u'2.2': -832918304}
        assert expected == c.hook_create_values()

    def test_out_version(self):
        expected_path = 'tests/sample-ceph-crushmap-compat.txt'
        out_path = expected_path + ".err"

        in_path = 'tests/sample-ceph-crushmap-compat.python-json'
        Ceph().main([
            'convert',
            '--in-path', in_path,
            '--out-path', out_path,
            '--out-format', 'txt',
            '--out-version', 'jewel',
        ])
        assert os.system("diff -Bbu " + expected_path + " " + out_path) == 0
        os.unlink(out_path)

        in_path = 'tests/sample-ceph-crushmap.python-json'
        with pytest.raises(Exception) as e:
            Ceph().main([
                'convert',
                '--in-path', in_path,
                '--out-path', out_path,
                '--out-format', 'txt',
                '--out-version', 'jewel',
            ])
        assert 'version lower than luminous' in str(e.value)

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -vv -s tests/test_ceph.py"
# End:



# -*- mode: python; coding: utf-8 -*-
#
# Copyright (C) 2017 <contact@redhat.com>
#
# Author: Loic Dachary <loic@dachary.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from crush.ceph import CephCrush


class TestConvert(object):

    def define_crushmap(self, host_count):
        crushmap = {
            "trees": [
                {
                    "type": "root",
                    "id": -1,
                    "name": "dc1",
                    "children": [],
                }
            ],
            "rules": {
                "firstn": [
                    ["take", "dc1"],
                    ["chooseleaf", "firstn", 0, "type", "host"],
                    ["emit"]
                ],
                "indep": [
                    ["take", "dc1"],
                    ["chooseleaf", "indep", 0, "type", "host"],
                    ["emit"]
                ],
            }
        }
        crushmap['trees'][0]['children'].extend([
            {
                "type": "host",
                "id": -(i + 2),
                "name": "host%d" % i,
                "weight": 3,
                "children": [
                    {"id": (2 * i), "name": "device%02d" % (2 * i), "weight": 1},
                    {"id": (2 * i + 1), "name": "device%02d" % (2 * i + 1), "weight": 2},
                ],
            } for i in range(0, host_count)
        ])
        return crushmap

    def test_ceph_version_compat(self):
        crushmap = self.define_crushmap(2)
        first_weight = 123
        second_weight = 456
        crushmap['choose_args'] = {
            0: [
                {
                    'bucket_id': -1,
                    'weight_set': [
                        [first_weight, second_weight],
                    ]
                }
            ]
        }
        c = CephCrush()
        c.parse(crushmap)
        c.ceph_version_compat()
        compat = c.get_crushmap()
        assert 'choose_args' not in compat
        assert compat['trees'][1]['name'] == 'dc1-target-weight'
        bucket = compat['trees'][0]
        assert bucket['children'][0]['weight'] == first_weight
        assert bucket['children'][1]['weight'] == second_weight

# Local Variables:
# compile-command: "cd .. ; tox -e py27 -- -vv -s tests/test_ceph_convert.py"
# End:
